#include "atlas_misc.h"
#ifndef PFYDIST
   #define PFYDIST 1
#endif
#ifndef PFXDIST
   #define PFXDIST 0
#endif
#ifndef PFADIST
   #define PFADIST 0
#endif

#if !defined(ATL_3DNow) && !defined(ATL_SSE1) && \
    (PFADIST != 0 || PFXDIST != 0 || PFYDIST != 0)
   #include "atlas_prefetch.h"
#endif

#if defined(ATL_3DNow) || defined(ATL_SSE1)
   #ifndef PFIY
      #define PFIY prefetchnta
   #endif
   #ifndef PFIX
      #define PFIX prefetchnta
   #endif
   #ifndef PFIA
      #ifdef ATL_3DNow
         #define PFIA prefetchw
      #else
         #define PFIA prefetcht0
      #endif
   #endif
#endif
/*
 * X & A are prefetched in M loop PF[A,X]DIST (in bytes) ahead
 */
#if PFADIST == 0                /* flag for no prefetch */
   #define prefA(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefA(mem) __asm__ __volatile__ \
         (Mstr(PFIA) " %0" : : "m" (*(((char *)(mem))+PFADIST)))
   #else
      #if PFLVL == 2
         #define prefA(mem) ATL_pfl2W(((char*)mem)+PFADIST)
      #else
         #define prefA(mem) ATL_pfl1W(((char*)mem)+PFADIST)
      #endif
   #endif
#endif
#if PFXDIST == 0                /* flag for no prefetch */
   #define prefX(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefX(mem) __asm__ __volatile__ \
         (Mstr(PFIX) " %0" : : "m" (*(((char *)(mem))+PFXDIST)))
   #else
      #if PFLVL == 2
         #define prefX(mem) ATL_pfl2R(((char*)mem)+PFXDIST)
      #else
         #define prefX(mem) ATL_pfl1R(((char*)mem)+PFXDIST)
      #endif
   #endif
#endif
/*
 * Y is prefetched in N-loop, and always fetches next NU piece
 */
#if PFYDIST == 0                /* flag for no prefetch */
   #define prefY(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefY(mem) \
         __asm__ __volatile__ (Mstr(PFIY) " %0" : : "m" (*(((char *)(mem)))))
   #else
      #if PFLVL == 2
         #define prefY(mem) ATL_pfl2R(mem)
      #else
         #define prefY(mem) ATL_pfl1R(mem)
      #endif
   #endif
#endif
#ifndef ATL_CINT
   #define ATL_CINT const int
#endif
#ifndef ATL_INT
   #define ATL_INT int
#endif
void ATL_UGERK(ATL_CINT M, ATL_CINT N, const double *X, const double *Y, 
               double *A, ATL_CINT lda)
/*
 * 64x1 unrolled r1_c.
 * Extracted from ATLAS/tune/blas/gemv/atlas-l2g.base
 */
{
   ATL_CINT N1=(N/1)*1, M64=(M/64)*64, lda1=lda*1;
   ATL_INT j;

   for (j=N1; j; j -= 1, A += lda1, Y += 1)
   {
      double *A0=A;
      const register double y0=Y[0];
      ATL_INT i;
      prefY(Y+1+1-1);
      for (i=0; i < M64; i += 64)
      {
         const register double x0=X[i+0], x1=X[i+1], x2=X[i+2], x3=X[i+3],
                               x4=X[i+4], x5=X[i+5], x6=X[i+6], x7=X[i+7],
                               x8=X[i+8], x9=X[i+9], x10=X[i+10], x11=X[i+11],
                               x12=X[i+12], x13=X[i+13], x14=X[i+14],
                               x15=X[i+15], x16=X[i+16], x17=X[i+17],
                               x18=X[i+18], x19=X[i+19], x20=X[i+20],
                               x21=X[i+21], x22=X[i+22], x23=X[i+23],
                               x24=X[i+24], x25=X[i+25], x26=X[i+26],
                               x27=X[i+27], x28=X[i+28], x29=X[i+29],
                               x30=X[i+30], x31=X[i+31], x32=X[i+32],
                               x33=X[i+33], x34=X[i+34], x35=X[i+35],
                               x36=X[i+36], x37=X[i+37], x38=X[i+38],
                               x39=X[i+39], x40=X[i+40], x41=X[i+41],
                               x42=X[i+42], x43=X[i+43], x44=X[i+44],
                               x45=X[i+45], x46=X[i+46], x47=X[i+47],
                               x48=X[i+48], x49=X[i+49], x50=X[i+50],
                               x51=X[i+51], x52=X[i+52], x53=X[i+53],
                               x54=X[i+54], x55=X[i+55], x56=X[i+56],
                               x57=X[i+57], x58=X[i+58], x59=X[i+59],
                               x60=X[i+60], x61=X[i+61], x62=X[i+62],
                               x63=X[i+63];

         prefX(X);
         A0[i+0] += y0 * x0;
         prefA(A0);
         A0[i+1] += y0 * x1;
         A0[i+2] += y0 * x2;
         A0[i+3] += y0 * x3;
         A0[i+4] += y0 * x4;
         A0[i+5] += y0 * x5;
         A0[i+6] += y0 * x6;
         A0[i+7] += y0 * x7;
         A0[i+8] += y0 * x8;
         prefA(A0);
         A0[i+9] += y0 * x9;
         A0[i+10] += y0 * x10;
         A0[i+11] += y0 * x11;
         A0[i+12] += y0 * x12;
         A0[i+13] += y0 * x13;
         A0[i+14] += y0 * x14;
         A0[i+15] += y0 * x15;
         A0[i+16] += y0 * x16;
         prefA(A0);
         A0[i+17] += y0 * x17;
         A0[i+18] += y0 * x18;
         A0[i+19] += y0 * x19;
         A0[i+20] += y0 * x20;
         A0[i+21] += y0 * x21;
         A0[i+22] += y0 * x22;
         A0[i+23] += y0 * x23;
         A0[i+24] += y0 * x24;
         prefA(A0);
         A0[i+25] += y0 * x25;
         A0[i+26] += y0 * x26;
         A0[i+27] += y0 * x27;
         A0[i+28] += y0 * x28;
         A0[i+29] += y0 * x29;
         A0[i+30] += y0 * x30;
         A0[i+31] += y0 * x31;
         A0[i+32] += y0 * x32;
         prefA(A0);
         A0[i+33] += y0 * x33;
         A0[i+34] += y0 * x34;
         A0[i+35] += y0 * x35;
         A0[i+36] += y0 * x36;
         A0[i+37] += y0 * x37;
         A0[i+38] += y0 * x38;
         A0[i+39] += y0 * x39;
         A0[i+40] += y0 * x40;
         prefA(A0);
         A0[i+41] += y0 * x41;
         A0[i+42] += y0 * x42;
         A0[i+43] += y0 * x43;
         A0[i+44] += y0 * x44;
         A0[i+45] += y0 * x45;
         A0[i+46] += y0 * x46;
         A0[i+47] += y0 * x47;
         A0[i+48] += y0 * x48;
         prefA(A0);
         A0[i+49] += y0 * x49;
         A0[i+50] += y0 * x50;
         A0[i+51] += y0 * x51;
         A0[i+52] += y0 * x52;
         A0[i+53] += y0 * x53;
         A0[i+54] += y0 * x54;
         A0[i+55] += y0 * x55;
         A0[i+56] += y0 * x56;
         prefA(A0);
         A0[i+57] += y0 * x57;
         A0[i+58] += y0 * x58;
         A0[i+59] += y0 * x59;
         A0[i+60] += y0 * x60;
         A0[i+61] += y0 * x61;
         A0[i+62] += y0 * x62;
         A0[i+63] += y0 * x63;
      }
      for (i=M64; i < M; i++)
      {
         const register double x0 = X[i];
         A0[i] += y0 * x0;
      }
   }
}
